/**
* Copyright (C) 2005 - Bull S.A.
*
* Jimys - J2EE is monitoring your system
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
* USA
*
* --------------------------------------------------------------------------
* $Id: Pump.java,v 1.4 2005/09/15 14:22:15 goebelg Exp $
* --------------------------------------------------------------------------
*/
package org.objectweb.jimys.monitoredSystem.pump;

import org.apache.xerces.parsers.DOMParser;
import org.objectweb.jimys.monitoredSystem.admin.Admin;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;

import javax.management.*;
import javax.management.timer.Timer;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import java.io.*;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.DataFormatException;
import java.util.zip.Deflater;
import java.util.zip.Inflater;

/**
 * The pump class
 * @author Georges Goebel, Marc Calvisi
 */
public class Pump implements PumpMBean, NotificationBroadcaster, Serializable, NotificationEmitter {
    /**
     * Event logger
     */
    transient private static Logger logger = Logger.getLogger(Pump.class.getName());

    /**
     *  Saves the register notification listeners of the clients
     */
    transient private Hashtable notificationlisteners = new Hashtable();

    /**
     *  The notification listener user for sending the data to the clients
     */
    transient private static EmergencyNotification eNoti = null;

    /**
     *  The timer used to transfer the data to the clients in regular time intervals
     */
    transient private Timer timer = new Timer();

    /**
     * The sample rate of the pump.
     * Default value = 5 seconds
     */
    private long sampleTime = 5000;

    /**
     * Saves a reference to all the started threads
     * Objectname of the probe -> Thread reference
     */
    private Hashtable threads = new Hashtable();

    /**
     *  ObjectName of the probes
     */
    private Collection probes = new Vector();

    /**
     *  Xml result document
     */
    private static Document xmlResult = null;

    /**
     * The number of collected probe values
     */
    private int numberCollected = 0;

    /**
     *  The maximum of collected probe values before a forced notification is done
     *  Permits to control the size of the user data in the jmx notification message
     */
    private int maximumCollected = 100;

    /**
     * The timer who is responsible for the Heartbeat message exchanged between the pump and the clients
     * This timer should have a smaller time than the other timer as it helps to detect errors and the
     * message exchanged is much smaller so less overhead
     */
    transient private Timer heartBeat = new Timer();

    /**
     * The heartbeat notification listener
     */
    transient private static HeartBeatNotification heatBeatNotification = null;

    /**
     * The time for the heartbeat timer
     */
    private long heartBeatSampleTime = 7000;

    /**
     * Indicates if the xml sent to the clients is compressed
     */
    private boolean isCompressing = false;

    /**
     * Default constructer, initialsing the timers and notificationlisteners
     */
    public Pump() {
        logger.log(Level.FINEST, "Constructing the Pump");
        if (eNoti == null) {
            eNoti = new EmergencyNotification();

            //timer.addNotificationListener(eNoti, null, "some handback object");
            timer.addNotificationListener(eNoti, null, null);
            timer.start();
            timer.addNotification("the type", "the type", "dd", new Date(), sampleTime);

            xmlResult = createDomDocument();
            //rootElement = createNewRootElement();
            xmlResult.appendChild(createNewRootElement(xmlResult));


        }

        if (heatBeatNotification == null) {
            heatBeatNotification = new HeartBeatNotification();

            //heartBeat.addNotificationListener(heatBeatNotification, null, "heartbeat");
            heartBeat.addNotificationListener(heatBeatNotification, null, null);
            heartBeat.start();
            heartBeat.addNotification("the type", "the type", "dd", new Date(), heartBeatSampleTime);
        }

        logger.log(Level.INFO, "Pump created");
    }

    /**
     * //TODO Implementation must be done
     * @return TOTO
     */
    public MBeanNotificationInfo[] getNotificationInfo() {
        return null;
    }

    /**
     * Adds a new notification listener from a client to the pump
     * @param arg0 The notification listener
     * @param arg1 A filter
     * @param arg2 Object
     * @throws IllegalArgumentException
     */
    public void addNotificationListener(NotificationListener arg0, NotificationFilter arg1, Object arg2) throws IllegalArgumentException {
        //System.out.println("Adding notification listener : " + arg2);
        logger.log(Level.INFO, "Adding notification listener : " + arg2);
        //      synchronized (notificationlisteners) {
        notificationlisteners.put(arg0, arg2);
        //        }
    }

    /**
     * Removes the notification listener from the list
     * @param arg0 The listener to remove
     * @throws ListenerNotFoundException
     */
    public void removeNotificationListener(NotificationListener arg0) throws ListenerNotFoundException {
        logger.log(Level.INFO, "Removing notification listener : " + arg0);
        notificationlisteners.remove(arg0);
    }

    /**
     * Gets the value for the maximum of collected values from the probes
     * @return The value
     */
    public int getMaximumCollected() {
        return maximumCollected;
    }

    /**
     * Sets a new value for the maximum of the collected items
     * @param maximumCollected The new value
     */
    public void setMaximumCollected(int maximumCollected) {
        this.maximumCollected = maximumCollected;
    }

    /**
     * The time of the notification timer which returns the data to the clients
     * @return The sample time
     */
    public Long getSampleTime() {
        return new Long(sampleTime);  //To change body of implemented methods use File | Settings | File Templates.
    }

    /**
     * Getter for the compressing variabe
     * @return is the result compressed
     */
    public boolean isCompressing() {
        return isCompressing;
    }


    public void modifyCompressing (Boolean compressing) {
        if (compressing != null) {
            isCompressing = compressing.booleanValue();
        }
    }

    public String getHeartBeat() {
        String finalXmlResult = null;

        synchronized (Pump.class){
            Document doc = createDomDocument();
            doc.appendChild(createNewAdminRootElement(doc));

            MBeanServer mbs = Admin.getMBeanServer();
            for (Iterator iterator = probes.iterator(); iterator.hasNext();) {
                ObjectName objectName = (ObjectName) iterator.next();
                //       System.out.println(objectName);
                Element admin = doc.createElement("ATTRIBUTE");
                admin.setAttribute("probe", objectName.toString());
                admin.setAttribute("time", Long.toString(new Date().getTime()));
                Boolean isAlive = null;
                String fullIdentifier = null;
                try {
                    isAlive = (Boolean) mbs.getAttribute(objectName, "Alive");
                    fullIdentifier = (String) mbs.getAttribute(objectName, "FullIdentifier");
                } catch (InstanceNotFoundException e) {
                    //e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                    logger.log(Level.SEVERE, e.getMessage());
                } catch (MBeanException e) {
                    //e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                    logger.log(Level.SEVERE, e.getMessage());
                } catch (ReflectionException e) {
                    //e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                    logger.log(Level.SEVERE, e.getMessage());
                } catch (AttributeNotFoundException e) {
                    //e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                    logger.log(Level.SEVERE, e.getMessage());
                }
                if (fullIdentifier == null) {
                    fullIdentifier = "Error";
                } else {
                    fullIdentifier = fullIdentifier.substring(0, fullIdentifier.lastIndexOf(";"));
                }
                admin.setAttribute("fullIdentifier", fullIdentifier);


                if (isAlive != null) {
                    if (isAlive.booleanValue()) {
                        // everything ok
                        admin.setAttribute("status", "ok");
                    } else {
                        // component failed
                        admin.setAttribute("status", "component error");
                    }

                } else {
                    // org.objectweb.jimys.monitoredSystem.probe error
                    admin.setAttribute("status", "probeError");
                }
                doc.getFirstChild().appendChild(admin);
            }
            try {
                finalXmlResult = convertToString(doc);
            } catch (TransformerException e) {
                //e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                logger.log(Level.SEVERE, e.getMessage());
            }

        }
        return finalXmlResult;  //To change body of implemented methods use File | Settings | File Templates.

    }

    public Collection getRegisteredProbes() {
        Collection probes = new ArrayList();
        Enumeration enume = threads.keys();
        while (enume.hasMoreElements()) {
            probes.add(enume.nextElement());
        }
        return probes;
    }

    /**
     * A new sample time for the timer which returns the data to the clients
     * @param sampleTime The new value
     */
    public void setSampleTime(Long sampleTime) {
        logger.log(Level.FINEST, "Setting new SampleTime : "+sampleTime);
        Vector vec = timer.getAllNotificationIDs();
        for (int i = 0; i < vec.size(); i++) {
            Integer o = (Integer) vec.elementAt(i);

            String notiMessage = timer.getNotificationMessage(o);
            String notiType = timer.getNotificationType(o);
            Object notiUserData = timer.getNotificationUserData(o);

            try {
                timer.removeNotification(o);
                Date timerTriggerAt = new Date((new Date()).getTime() + 5000L);
                timer.addNotification(notiType, notiMessage, notiUserData, timerTriggerAt, sampleTime.longValue());
                this.sampleTime = sampleTime.longValue();
            } catch (InstanceNotFoundException e) {
                logger.log(Level.SEVERE, e.getMessage());
                //e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
            }
            //System.out.println("oo " + notiMessage + " " + notiType + " " + notiUserData);
            logger.log(Level.FINEST, "Setting new SampleTime Timer restarted ");
        }
    }

    /**
     * Get the sample rate of a probe
     * @param probename The objectname in string format
     * @return The sample rate of the pump. Null if not found
     */
    public Long getProbeSampleTime(String probename) {
        logger.log(Level.FINEST, "getProbeSampleTime  : "+probename);
        RunningThread rt = (RunningThread) this.threads.get(probename);
        if (rt != null) {
            return new Long(rt.getSampleTime());
        }
        return null;  //To change body of implemented methods use File | Settings | File Templates.
    }

    /**
     * Set a new Sample rate a probe
     * @param probename The probane (ObjectName)
     * @param newSampleTime The new sample rate
     */
    public void setProbeSampleTime(String probename, Long newSampleTime) {
        logger.log(Level.FINEST, "setProbeSampleTime  : "+probename + " new sampletime : "+newSampleTime);
        RunningThread rt = (RunningThread) this.threads.get(probename);
        if ((rt != null) && (newSampleTime != null)) {
            rt.setSampleTime(newSampleTime.longValue());
        }
    }

    /**
     * Adds a new notification to the pump
     * @param mBeanName
     * @throws MalformedObjectNameException
     * @throws InstanceNotFoundException
     */
    public void addEmergencyProbeNotification(String mBeanName) throws MalformedObjectNameException, InstanceNotFoundException {
        ObjectName delegateName = ObjectName.getInstance(mBeanName);
        logger.log(Level.FINEST, "addEmergencyProbeNotification  : "+mBeanName);
        //System.out.println("MBean : " + delegateName + " eNoti : " + eNoti);
        MBeanServer mbs = Admin.getMBeanServer();
        mbs.addNotificationListener(delegateName, eNoti, null, "");

    }

    /**
     * Starts a new probe in the pump
     * @param mBeanName The objectname of the probe
     * @param attributes The attributes of the MBean to observe
     * @param probeSampleTime The probe sample time
     * @throws MalformedObjectNameException
     */
    public void runMonitoringOnProbe(String mBeanName, Collection attributes, Long probeSampleTime) throws MalformedObjectNameException {
        logger.log(Level.FINEST, "runMonitoringOnProbe  : "+mBeanName+ " attributes "+attributes + " probeSampleTime "+probeSampleTime);
        ObjectName delegateName = ObjectName.getInstance(mBeanName);

        long pSampleT = sampleTime;
        if (probeSampleTime != null) {
            pSampleT = probeSampleTime.longValue();
        }
                                                    //TODO
        RunningThread newThread = new RunningThread(delegateName, attributes, this, pSampleT);
        newThread.start();

        threads.put(mBeanName, newThread);
        probes.add(delegateName);
    }

    /**
     * The probes (Threads) use this method to add collected data to the pump
     * The collection of the data must have the same size than the collection of the attributenames
     * The format of tha data inside the collection are objects which are evaluated to string
     * The object in the collection can also be a xml stream as xml for complex data
     * @param ressource The resource name (xml tag)
     * @param ID id of the probe (mbean name)
     * @param attributeName the attributes which are observed
     * @param data The collection of the data
     */
    public synchronized void addData(String ressource, String ID, Collection attributeName, Collection data) {
        logger.log(Level.FINEST, "Pump addData  : "+ ressource + " data "+data);


        try {
            synchronized (Pump.class) {
                Element elem = xmlResult.createElement(ressource);
                elem.setAttribute("id", ID);
                elem.setAttribute("time", new Long(new Date().getTime()).toString());

                int length = attributeName.size();
                Iterator itAttributeName = attributeName.iterator();
                Iterator itData = data.iterator();

                if (data.size() != attributeName.size()) {
                    return;
                }

                for (int i = 0; i < length; i++) {
                    String aName = (String) itAttributeName.next();
                    Object oData = itData.next();

                    if (oData.toString().indexOf("<?xml version=\"1.0\" encoding=\"UTF-8\"?>") == 0) {
                        Node parsedData = parseXml(oData.toString());
                        Node imported = xmlResult.importNode(parsedData, true);
                        //NodeList nl = imported.getChildNodes();

                        Node firstChild = imported.getFirstChild();
                        while (firstChild != null) {
                            elem.appendChild(firstChild.cloneNode(true));
                            firstChild = firstChild.getNextSibling();
                        }


                    } else {
                        Element attribute = xmlResult.createElement("ATTRIBUTE");
                        attribute.setAttribute("name", aName);
                        //attribute.setAttribute("value",oData.toString());
                        Node value = xmlResult.createTextNode(oData.toString());
                        attribute.appendChild(value);
                        elem.appendChild(attribute);
                    }


                    if (xmlResult == null) logger.log(Level.SEVERE, "Pump addData  : xmlResult == null");//System.out.println("            : xmlResult == null");
                    if (xmlResult.getFirstChild() == null) logger.log(Level.SEVERE, "Pump addData  : mlResult.getFirstChild() == null");//System.out.println("            : xmlResult.getFirstChild() == null");
                    //rootElement.appendChild(elem);
                    xmlResult.getFirstChild().appendChild(elem);
                }

                numberCollected++;
                if (numberCollected > maximumCollected) {
                    timer.sendNotification(new Notification("manual Notification", "forced notification", 9000));
                    logger.log(Level.FINEST, "Manual send notification !!!");
                }
            }
        } catch (Exception e) {
            //e.printStackTrace();
            logger.log(Level.SEVERE, e.getMessage());
            System.exit(0);
        }
    }

    /**
     * Stops all the probes (1 thread for each probe)
     */
    public void stopAllThreads() {
        //System.out.println("Stopping all threads");
        logger.log(Level.INFO, "stopAllThreads");
        Collection col = threads.values();
        for (Iterator iterator = col.iterator(); iterator.hasNext();) {
            RunningThread runningThread = (RunningThread) iterator.next();
            runningThread.stopMe();
            logger.log(Level.FINEST, "stopAllThreads : "+runningThread.getObjectName());
        }
    }

    /**
     * Stops one probe (one thread)
     * @param objectName The objectname of the probe
     */
    public void stopThread(String objectName) {
        logger.log(Level.INFO, "stopThread : "+objectName);
        RunningThread rThread = (RunningThread) threads.get(objectName.toString());
        if (rThread != null) {
            rThread.stopMe();
        } else {
            logger.log(Level.WARNING, "stopThread : No such thread started "+objectName);
            //System.out.println("No such thread started");
        }
    }

    public void removeNotificationListener(NotificationListener notificationListener, NotificationFilter notificationFilter, Object object) throws ListenerNotFoundException {
        synchronized(notificationlisteners) {
            Object o = notificationlisteners.remove(notificationListener);
            logger.log(Level.INFO, "Removed the listener : "+o);
        }
    }

    /**
     * The Thread which is started for each probe
     */
    private class RunningThread extends Thread implements Serializable {
        /**
         * The reference to the mbean server
         */
        transient private MBeanServer mbs = null;

        /**
         * The reference to the pump
         */
        transient private Pump father = null;

        /**
         * Object name of the probe
         */
        private ObjectName oName = null;

        /**
         * The attributes to observe
         */
        transient private Collection attributes = null;

        /**
         * The samplerate of the probe
         */
        private long sampleTime;

        /**
         * Used to stop the thread
         */
        private boolean runme = true;

        /**
         * Constructor for the initialisation of the thread
         * @param oName Objectname of the probe
         * @param attributes Attributes to observe for the mbean
         * @param father The reference to the pump
         * @param sampleTime The sample time for the probe
         */
        public RunningThread(ObjectName oName, Collection attributes, Pump father, long sampleTime) {
            logger.log(Level.FINEST, "constructor : RunningThread : "+oName, " sample rate : " +sampleTime + " father "+father);
            this.oName = oName;
            this.attributes = attributes;
            this.father = father;
            this.sampleTime = sampleTime;

            mbs = Admin.getMBeanServer();
        }

        /**
         * Synchronized run methode
         */
        public synchronized void run() {
            //System.out.println("Running thread with : " + oName + " and sampleTime : " + sampleTime);
            logger.log(Level.INFO, "Running thread with : " + oName + " and sampleTime : " + sampleTime);
            try {
                while (runme) {
                    this.wait(sampleTime);

                    String probeID = null;

                    try {
                        probeID = (String) mbs.getAttribute(oName, "Identifier");

                    } catch (AttributeNotFoundException e) {
                        //System.out.println("   -- Attribute : Identifier not found");
                        logger.log(Level.WARNING, "-- Attribute : Identifier not found "+oName);
                    }

                    Collection data = new ArrayList();
                    if (probeID == null) probeID = "unknown";

                    /*
                    String result = getHostName();
                    result += ";" + "domain" + ";" + "instance" + ";" + "probe" + ";" + probeID;
                    */
                    String fullIdentifier = (String) mbs.getAttribute(oName, "FullIdentifier");

                    for (Iterator iterator = attributes.iterator(); iterator.hasNext();) {
                        String attribute = (String) iterator.next();
                        Object o = mbs.getAttribute(oName, attribute);
                        if (o != null) {
                            //   System.out.println("attribute " + attribute + " is : " + o.toString());
                            //System.out.println("attribute " + attribute);
                            logger.log(Level.FINEST, oName + " attribute "+attribute);
                            data.add(o);


                        }
                    }
                    // TODO Change to fullIdentifier    .substring(0, fullIdentifier.lastIndexOf(";"))
                    fullIdentifier = fullIdentifier.substring(0, fullIdentifier.lastIndexOf(";"));
                    father.addData(probeID, fullIdentifier+";"+oName.toString(), attributes, data);
                }
            } catch (InterruptedException e) {
                //e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                logger.log(Level.SEVERE, e.getMessage());
            } catch (ReflectionException e) {
                //e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                logger.log(Level.SEVERE, e.getMessage());
            } catch (InstanceNotFoundException e) {
                //e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                logger.log(Level.SEVERE, e.getMessage());
            } catch (MBeanException e) {
                //e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                logger.log(Level.SEVERE, e.getMessage());
            } catch (AttributeNotFoundException e) {
                //e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                logger.log(Level.SEVERE, e.getMessage());
            }

        }

        /**
         * Getter for the sample time
         * @return The sampletime
         */
        public long getSampleTime() {
            return sampleTime;
        }

        /**
         * Setter for the sample time
         * @param sampleTime The new sampletime
         */
        public void setSampleTime(long sampleTime) {
            this.sampleTime = sampleTime;
        }

        /**
         * Getter for the stop variable
         * @return stop value
         */
        public boolean isStop() {
            return runme;
        }

        /**
         * Stops the thread
         */
        public void stopMe() {
            this.runme = false;
        }

        /**
         * Getter for the objectname of the probe
         * @return The objectname
         */
        public ObjectName getObjectName() {
            return oName;
        }
    }

    /**
     * The class used fot the notificationlistener to send the data to the clients
     */
    private class EmergencyNotification implements NotificationListener, Serializable {

        /**
         * Executed when the timer has expired or when a forced notification is done
         * @param notification The notification object
         * @param object Object
         */
        public synchronized void handleNotification(Notification notification, Object object) {
            //System.out.println("Getting emergency Event ! : " + notification.getUserData());
            logger.log(Level.FINEST, "Getting EmergencyNotification");
            //TODO Implement the list for notifications
            if (xmlResult != null) {
                synchronized (Pump.class) {
                    //                    xmlResult.appendChild(rootElement);

                    try {
                        String finalXmlResult = convertToString(xmlResult);

                        logger.log(Level.FINEST, "RESULT SIZE Uncompressed:" + finalXmlResult.getBytes().length);
                        logger.log(Level.FINEST, "RESULT SIZE Compressed:" + compress(finalXmlResult).length);

                        xmlResult = createDomDocument();
                        xmlResult.appendChild(createNewRootElement(xmlResult));
                        Enumeration enumer = notificationlisteners.keys();
                        while (enumer.hasMoreElements()) {
                            NotificationListener noti = (NotificationListener) enumer.nextElement();
                            Object arg2 = notificationlisteners.get(noti);
                            long now = System.currentTimeMillis();
                            Notification newNoti = new Notification("mbean=org.objectweb.jimys.monitoredSystem.pump", this, now, now, "Message");
                            if (isCompressing()){
                                newNoti.setUserData(compress(finalXmlResult));
                            } else {
                                newNoti.setUserData(finalXmlResult);
                            }
                            noti.handleNotification(newNoti, arg2);
                        }
                        // TODO CHECK IF CORRECT
                        numberCollected = 0;
                    } catch (TransformerException e) {
                        //e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                        logger.log(Level.SEVERE, e.getMessage());
                    }

                }
            }
        }
    }

    /**
     * The class used fot the notificationlistener to send the heartbeat to the clients
     */
    private class HeartBeatNotification implements NotificationListener, Serializable {

        /**
         * Executed when the timer has expired or when a forced notification is done
         * @param notification The notification object
         * @param object Object
         */
        public synchronized void handleNotification(Notification notification, Object object) {
            //System.out.println("Getting heartbeat Event ! : " + notification.getUserData());
            logger.log(Level.FINEST, "Getting HeartBeatNotification");
            synchronized (Pump.class) {

                Document doc = createDomDocument();
                doc.appendChild(createNewAdminRootElement(doc));

                MBeanServer mbs = Admin.getMBeanServer();
                for (Iterator iterator = probes.iterator(); iterator.hasNext();) {
                    ObjectName objectName = (ObjectName) iterator.next();
                    //       System.out.println(objectName);
                    Element admin = doc.createElement("ATTRIBUTE");
                    admin.setAttribute("probe", objectName.toString());
                    admin.setAttribute("time", Long.toString(new Date().getTime()));
                    Boolean isAlive = null;
                    String fullIdentifier = null;
                    try {
                        isAlive = (Boolean) mbs.getAttribute(objectName, "Alive");
                        fullIdentifier = (String) mbs.getAttribute(objectName, "FullIdentifier");
                    } catch (InstanceNotFoundException e) {
                        //e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                        logger.log(Level.SEVERE, e.getMessage());
                    } catch (MBeanException e) {
                        //e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                        logger.log(Level.SEVERE, e.getMessage());
                    } catch (ReflectionException e) {
                        //e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                        logger.log(Level.SEVERE, e.getMessage());
                    } catch (AttributeNotFoundException e) {
                        //e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                        logger.log(Level.SEVERE, e.getMessage());
                    }
                    if (fullIdentifier == null) {
                        fullIdentifier = "Error";
                    } else {
                        fullIdentifier = fullIdentifier.substring(0, fullIdentifier.lastIndexOf(";"));
                    }
                    admin.setAttribute("fullIdentifier", fullIdentifier);


                    if (isAlive != null) {
                        if (isAlive.booleanValue()) {
                            // everything ok
                            admin.setAttribute("status", "ok");
                        } else {
                            // component failed
                            admin.setAttribute("status", "component error");
                        }

                    } else {
                        // org.objectweb.jimys.monitoredSystem.probe error
                        admin.setAttribute("status", "probeError");
                    }
                    doc.getFirstChild().appendChild(admin);
                }
                String finalXmlResult = null;
                try {
                    finalXmlResult = convertToString(doc);
                } catch (TransformerException e) {
                    //e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                    logger.log(Level.SEVERE, e.getMessage());
                }


                Enumeration enumer = notificationlisteners.keys();
                while (enumer.hasMoreElements()) {
                    NotificationListener noti = (NotificationListener) enumer.nextElement();
                    Object arg2 = notificationlisteners.get(noti);
                    long now = System.currentTimeMillis();
                    Notification newNoti = new Notification("mbean=org.objectweb.jimys.monitoredSystem.pump", this, now, now, "Message");
                    if (isCompressing) {
                        newNoti.setUserData(compress(finalXmlResult));
                    } else {
                        newNoti.setUserData(finalXmlResult);
                    }

                    noti.handleNotification(newNoti, arg2);
                }

            }

        }
    }

    /**
     * Getter for the hostname of the machine
     * @return The hostname of the machine
     */
    public synchronized static String getHostName() {
        try {
            // Get hostname by textual representation of IP address
            InetAddress addr = InetAddress.getLocalHost();

            // Get hostname
            String hostname = addr.getHostName();
            return hostname;
        } catch (UnknownHostException e) {
            //e.printStackTrace();
            logger.log(Level.SEVERE, e.getMessage());
        }
        return "Null : hostname";
    }

    /**
     * Create a new xml document
     * @return The new xml document
     */
    public synchronized static Document createDomDocument() {
        try {
            DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
            Document doc = builder.newDocument();
            return doc;
        } catch (ParserConfigurationException e) {
            logger.log(Level.SEVERE, e.getMessage());
        }
        return null;
    }

    /**
     * Create a new root node for a given document used to send the data
     * @param doc The xml document
     * @return The new root element
     */
    private synchronized static Element createNewRootElement(Document doc) {
        Element rootElement = doc.createElement("pump");

        String pumpId = getHostName();
        pumpId += ";" + "domain" + ";" + "instance" + ";" + "pump" + ";" + "PUMP";

        rootElement.setAttribute("pumpID", pumpId);
        rootElement.setAttribute("createTime", new Long(new Date().getTime()).toString());

        return rootElement;
    }

    /**
     * Create a new root node for a given document used to send the heartbeat
     * @param doc The xml document
     * @return The new root element
     */
    private synchronized static Element createNewAdminRootElement(Document doc) {
        Element rootElement = doc.createElement("admin");

        String adminId = getHostName();
        adminId += ";" + "domain" + ";" + "instance" + ";" + "admin" + ";" + "CONTROL";

        rootElement.setAttribute("adminId", adminId);
        rootElement.setAttribute("createTime", new Long(new Date().getTime()).toString());

        return rootElement;
    }

    /**
     * Convert an xml document to a string
     * @param doc the xml document
     * @return The string
     * @throws TransformerException
     */
    private synchronized static String convertToString(Document doc) throws TransformerException {
        TransformerFactory tFactory = TransformerFactory.newInstance();
        Transformer transformer = tFactory.newTransformer();

        DOMSource source = new DOMSource(doc);
        StringWriter sw = new StringWriter();
        StreamResult result = new StreamResult(sw);
        transformer.transform(source, result);
        String xmlString = sw.toString();
        return xmlString;
    }

    /**
     * Parse an string to an xml node
     * @param s The xml string
     * @return The xml node
     */
    public synchronized static Node parseXml(String s) {
        DOMParser d = new DOMParser();
        try {
            StringReader sr = new StringReader(s);
            InputSource is = new InputSource(sr);
            d.parse(is);
            Node node = d.getDocument().getFirstChild();
            sr.close();
            return node;
        } catch (Exception e) {
            e.printStackTrace();
            //System.err.println("No XML Format " + s);
            logger.log(Level.SEVERE, "XML DATA " +s + '\n' + e.getMessage());
        }
        return null;
    }


    /**
     * Combress the string send to the clients
     * @param data The string to compress
     * @return The compressed string in a byte array
     */
    public static byte[] compress(String data) {
        if (data == null) return null;

        Deflater compressor = new Deflater();
        compressor.setLevel(Deflater.BEST_COMPRESSION);

        byte[] input = data.getBytes();
        // Give the compressor the data to compress
        compressor.setInput(input);
        compressor.finish();

        // Create an expandable byte array to hold the compressed data.
        // You cannot use an array that's the same size as the orginal because
        // there is no guarantee that the compressed data will be smaller than
        // the uncompressed data.
        ByteArrayOutputStream bos = new ByteArrayOutputStream(input.length);

        // Compress the data
        byte[] buf = new byte[1024];
        while (!compressor.finished()) {
            int count = compressor.deflate(buf);
            bos.write(buf, 0, count);
        }
        try {
            bos.close();
        } catch (IOException e) {
            logger.log(Level.SEVERE, e.getMessage());
        }

        // Get the compressed data
        byte[] compressedData = bos.toByteArray();
        return compressedData;
    }

    /**
     * Decompress a byte array to a string
     * @param data The compressed data
     * @return The decompressed data in a string
     */
    public static String decompress(byte[] data) {
        if (data == null) return null;

        Inflater decompressor = new Inflater();
        decompressor.setInput(data);

        // Create an expandable byte array to hold the decompressed data
        ByteArrayOutputStream bos2 = new ByteArrayOutputStream(data.length);

        // Decompress the data
        byte[] buf2 = new byte[1024];
        while (!decompressor.finished()) {
            try {
                int count = decompressor.inflate(buf2);
                bos2.write(buf2, 0, count);
            } catch (DataFormatException e) {
                logger.log(Level.SEVERE, e.getMessage());
            }
        }
        try {
            bos2.close();
        } catch (IOException e) {
            logger.log(Level.SEVERE, e.getMessage());
        }

        // Get the decompressed data
        byte[] decompressedData = bos2.toByteArray();
        return new String(decompressedData);
    }

}
